function [delay_samples,delay_ms]=delay_by_excitation(speech_frame_1,speech_frame_2,fs_kHz,plot_flag);

%--------------------------------------------------------------------------
% This function computes the time delay between two speech signals 
% using the excitation source information.
%--------------------------------------------------------------------------
% REFERENCE:
%--------------------------------------------------------------------------
%
% 1. Tracking a moving speaker using excitation source information 
%    Vikas C. Raykar, B.Yegnanarayana, S. R. Mahadeva Prasanna, and, 
%    Ramani Duraiswami (Eurospeech 2003, Geneva, September 2003) 
%
% 2. Speaker Localization using excitation source information in speech 
%    Vikas C. Raykar, B.Yegnanarayana, S. R. Mahadeva Prasanna, and,
%    Ramani Duraiswami
%    (accepted for IEEE Transactions on Speech and Audio Processing) 
%
%  These two papers can be downloaded from:
%  http://cvl.umiacs.umd.edu/users/vikas/publications/papers.html
%--------------------------------------------------------------------------
% INPUT:
%--------------------------------------------------------------------------
% speech_frame_1      :Speech segment from the first  mic
% speech_frame_1      :Speech segment from the second mic
% fs_kHz              :Sampling rate in KHz
% plot_flag           :if 1 plotting is enabled      
%--------------------------------------------------------------------------
% OUTPUT:
%--------------------------------------------------------------------------
% delay_samples       : vector of delays in number of samples
% delay_ms            : vector of delays in milliseconds
%--------------------------------------------------------------------------
% AUTHORS   : Vikas Chandrakant Raykar and S. R. Mahadeva Prasanna
% CONTACT   : vikas@umiacs.umd.edu
% DATE      : 11 August 2004
%--------------------------------------------------------------------------
% NOTE:
%--------------------------------------------------------------------------
% 1. If you wish you can change the frame size and the frame shift and
%    the order used to compute the linear prediction residual in the
%    function.
%--------------------------------------------------------------------------

%Make sure both the frames are of same length

N=min(length(speech_frame_1),length(speech_frame_2));
speech_frame_1=speech_frame_1(1:N);
speech_frame_2=speech_frame_2(1:N);

%Compute the LP residual for both the frames

LP_frame_size=ceil(25*fs_kHz); %20 ms 
LP_frame_shift=ceil(1*fs_kHz); %1  ms
LP_order=10;
preempflag=0;

r1=LP_Residual(speech_frame_1,LP_frame_size,LP_frame_shift,LP_order,preempflag);
r2=LP_Residual(speech_frame_2,LP_frame_size,LP_frame_shift,LP_order,preempflag);

%Compute the hilbert envelope of the LP resiudals

h1=henv(real(r1)); ah1=rem_trend(h1,fs_kHz);
h2=henv(real(r2)); ah2=rem_trend(h2,fs_kHz);

%Compute the cross correlation of the hilbert envelope of the LP residuals

y=xcorr(ah1-mean(ah1),ah2-mean(ah2));

%Find the location of the peak in the cross correlation function

[max_y,max_index]=max(y);
delay_samples=max_index-N;  
delay_ms=delay_samples/fs_kHz;


if plot_flag==1
    figure;
    time_ms=[0:1/fs_kHz:(N-1)/fs_kHz];
    
    
    subplot(4,2,1);
    plot(time_ms,speech_frame_1,'r');
    min_y=min([min(speech_frame_1),min(speech_frame_2)]);
    max_y=max([max(speech_frame_1),max(speech_frame_2)]);
    axis([time_ms(1) time_ms(end) min_y max_y]);
    xlabel('Time (ms)');
    title('Speech Frame 1');
    ylabel('Speech Signal');
    
    subplot(4,2,2);
    plot(time_ms,speech_frame_2,'r');
    axis([time_ms(1) time_ms(end) min_y max_y]);
    xlabel('Time (ms)');
    title('Speech Frame 2');
    ylabel('Speech Signal');
    
    subplot(4,2,3);
    plot(time_ms,r1,'b');
    min_y=min([min(r1),min(r2)]);
    max_y=max([max(r1),max(r2)]);
    axis([time_ms(1) time_ms(end) min_y max_y]);
    xlabel('Time (ms)');
    ylabel('LP Residual');
    
    subplot(4,2,4);
    plot(time_ms,r2,'b');
    axis([time_ms(1) time_ms(end) min_y max_y]);
    xlabel('Time (ms)');
    ylabel('LP Residual');
    
    subplot(4,2,5);
    plot(time_ms,ah1,'k');
    min_y=min([min(ah1),min(ah2)]);
    max_y=max([max(ah1),max(ah2)]);
    axis([time_ms(1) time_ms(end) min_y max_y]);
    xlabel('Time (ms)');
    ylabel('Hilbert Envelope');
    
    subplot(4,2,6);
    plot(time_ms,ah2,'k');
    axis([time_ms(1) time_ms(end) min_y max_y]);
    xlabel('Time (ms)');
    ylabel('Hilbert Envelope');
    
    subplot(4,1,4);
    h=plot(y,'k');
    set(h,'linewidth',2);
    axis([1 length(y) min(y) max(y)]);
    title('Cross correlation fucntion');
    hold on;
    X=[N N];
    Y=[min(y) max(y)];
    h=line(X,Y);
    set(h,'linestyle',':','color','k');
    hold on;
    X=[N+delay_samples N+delay_samples];
    Y=[min(y) max(y)];
    h=line(X,Y);
    set(h,'linestyle','-','color','r','linewidth',1);
    legend(sprintf('Delay = %d samples %f ms',delay_samples,delay_ms));
    
    
end

%--------------------------------------------------------------------------
% Function to compute the LP residual
%--------------------------------------------------------------------------

function [residual,LPCoeffs] = LP_Residual(speech,framesize,frameshift,lporder,preempflag)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUTS :
%		speech - speech in ASCII
%		framesize - framesize for lpanalysis (samples)
%		frameshift - frameshift for lpanalysis (samples)
%		lporder    - order of lpc
%		preempflag - If 0 do preemphasis
%		plotflag  - If 1 plot results
%  OUTPUTS :
%		residual - residual signal
%		LPCoeffs - 2D array containing LP coeffs of all frames
%  This matlab function generates LPresidual by inverse filtering. The inverse
%  filter coefficients are obtained by LPanalysis.
% NOTE: To eliminate (lporder+1) poorly predicted samples in the begining of 
% each frame, we need to consider overlapping frames for analysis.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% METHOD: 
%	-Consider one frame at a time, compute LP coefficients, pass orginal
%        frame thro' inverse filter to get corresponding reisdual frame.
%	-Consider only FRAMESHIFT number of samples among the total 
%        FRAMESIZE number of samples.
%	-Repeat same procedure for all the frames.
%	-In the end, hamming window lporder number of 
%        samples of the first frame 
%	 to eliminate initial poorly predicted samples.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% AUTHOR : Mahadeva Prasanna S.R.
% DATE   : 01-07-2000
% LAST MODIFIED : 01-07-2000, 22-10-2000, 24-10-2000, 7-10-2002
% OTHER FUNCTIONS CALLED :
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% PREEMPHASIZING SPEECH SIGNAL

if(preempflag==0)
    dspeech=diff(speech);
    dspeech(length(dspeech)+1)=dspeech(length(dspeech));
else
    dspeech=speech;
end

speech=dspeech;

%COMPUTING RESIDUAL

residual = zeros(1,length(speech));

%NUMBER OF FRAMES

nframes=floor((length(speech)-framesize)/frameshift)+1;

LPCoeffs=zeros(nframes,(lporder+1));

j=1;
for i=1:frameshift:(length(speech)-framesize)
    
    SpFrm = speech(i:i+framesize-1);
    
    lpcoef = lpc(hamming(length(SpFrm)).*SpFrm,lporder);
    
    LPCoeffs(j,:)=real(lpcoef);
    
    if(i<=lporder)
        
        PrevFrm=zeros(1,lporder);
        
    else
        PrevFrm=speech((i-lporder):(i-1));
    end
    
    
    ResFrm=ResFilter(real(PrevFrm),real(SpFrm),real(lpcoef),lporder,framesize);
    
    j=j+1;
    
    residual(i:i+frameshift-1)=ResFrm(1:frameshift);
    
end

residual(i+frameshift:i+framesize-1)=ResFrm(frameshift+1:framesize);

%PROCESSING LASTFRAME SAMPLES

if(i<length(speech))
    
    SpFrm = speech(i:length(speech));
    
    lpcoef = lpc(hamming(length(SpFrm)).*SpFrm,lporder);
    
    LPCoeffs(j,:)=real(lpcoef);
    
    PrevFrm=speech((i-lporder):(i-1));
    
    ResFrm=ResFilter(real(PrevFrm),real(SpFrm),real(lpcoef),lporder,framesize);
    
    j=j+1;
    
    residual(i:i+length(ResFrm)-1)=ResFrm(1:length(ResFrm));
    
end

hm=hamming(2*lporder);

for(i=1:length(hm)/2)
    
    residual(i)=residual(i)*hm(i);      %attenuating first lporder samples
    
end


function [ResFrm]=ResFilter(PrevSpFrm,SpFrm,FrmLPC,LPorder,FrmSize);

ResFrm=zeros(1,FrmSize);

tempfrm=zeros(1,FrmSize+LPorder);

tempfrm(1:LPorder)=PrevSpFrm;

tempfrm(LPorder+1:LPorder+FrmSize)=SpFrm(1:FrmSize);


for(i=1:FrmSize)
    
    t=0;
    for(j=1:LPorder)
        
        t=t+FrmLPC(j+1)*tempfrm(-j+i+LPorder);
    end
    
    
    ResFrm(i)=SpFrm(i)-(-t);
    
end


%--------------------------------------------------------------------------
% Function to compute the hilbert envelope
%--------------------------------------------------------------------------

function [hilbertenv]=henv(signal)

tempseq=hilbert(signal);

% returns a complex helical sequence, sometimes called the analytic
% signal, from a real data sequence. The analytic signal has a real
% part, which is the original data, and an imaginary part, which
% contains the Hilbert Transform. The imaginary part is a version 
% of the original real sequence with a 90 degree pahse shift. Sines
% are therefore transformed to cosines and vice versa.

hilbertseq=imag(tempseq);

% hilbertseq contains the hilbert transformed version of input signal
% hilb env is given by t=sqrt((res*res)+(h(res)*h(res)));

sigsqr=signal.*signal;

hilbsqr=hilbertseq.*hilbertseq;

henv=sqrt(sigsqr+hilbsqr);

hilbertenv=henv;

%--------------------------------------------------------------------------
% Function to remove the trend in the hilbert envelope
%--------------------------------------------------------------------------

function [Ret] = rem_trend(h,fs_kHz);

pd = floor(fs_kHz*2/1000);

for i = 1:pd
    temp = h(1:i+pd);
    Ret(i) = h(i)/mean(temp);
    Ret(i) = Ret(i)*h(i);
end;

for i = pd+1:length(h)-pd
    temp1 = h(i-pd:i+pd);
    Ret(i) = h(i)/mean(temp1);
    Ret(i) = Ret(i)*h(i);
end;

for i = (length(h)-pd+1):length(h)
    temp2 = h(i:length(h));
    Ret(i) = h(i)/mean(temp2);
    Ret(i) = Ret(i)*h(i);
end;

